#include <gst/video/gstvideofilter.h>

#include <string.h>
#include <math.h>

#include <gst/video/video.h>

#define GST_TYPE_REDGREENTV \
  (gst_redgreentv_get_type())
#define GST_REDGREENTV(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_REDGREENTV,GstRedGreenTV))
#define GST_REDGREENTV_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_REDGREENTV,GstRedGreenTVClass))
#define GST_IS_REDGREENTV(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_REDGREENTV))
#define GST_IS_REDGREENTV_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_REDGREENTV))

typedef struct _GstRedGreenTV GstRedGreenTV;
typedef struct _GstRedGreenTVClass GstRedGreenTVClass;

#define LAST_NEUTRAL 0
#define LAST_GREEN 1
#define LAST_RED 2

#define DEFAULT_RED_WEIGHT 45.0
#define DEFAULT_GREEN_WEIGHT 50.0

enum
{
  ARG_0,
  ARG_RED_WEIGHT,
  ARG_GREEN_WEIGHT,
  ARG_DIRECTION
};

struct _GstRedGreenTV
{
  GstVideoFilter videofilter;

  gint width, height;

  gdouble red_weight, green_weight;
  gdouble direction;
};

struct _GstRedGreenTVClass
{
  GstVideoFilterClass parent_class;
};

GType gst_redgreentv_get_type (void);

static const GstElementDetails redgreentv_details =
GST_ELEMENT_DETAILS ("RedGreenTV effect",
    "Filter/Effect/Video",
    "RedGreenTV does realtime goo'ing of the video input",
    "Sam Lantinga <slouken@devolution.com>");

static GstStaticPadTemplate gst_redgreentv_src_template =
    GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_RGBx)
    );

static GstStaticPadTemplate gst_redgreentv_sink_template =
    GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_RGBx)
    );

static GstVideoFilterClass *parent_class = NULL;

static gboolean
gst_redgreentv_set_caps (GstBaseTransform * btrans, GstCaps * incaps,
    GstCaps * outcaps)
{
  GstRedGreenTV *filter = GST_REDGREENTV (btrans);
  GstStructure *structure;
  gboolean ret = FALSE;

  structure = gst_caps_get_structure (incaps, 0);

  if (gst_structure_get_int (structure, "width", &filter->width) &&
      gst_structure_get_int (structure, "height", &filter->height)) {
    ret = TRUE;
  }

  return ret;
}

static gboolean
gst_redgreentv_get_unit_size (GstBaseTransform * btrans, GstCaps * caps,
    guint * size)
{
  GstRedGreenTV *filter;
  GstStructure *structure;
  gboolean ret = FALSE;
  gint width, height;

  filter = GST_REDGREENTV (btrans);

  structure = gst_caps_get_structure (caps, 0);

  if (gst_structure_get_int (structure, "width", &width) &&
      gst_structure_get_int (structure, "height", &height)) {
    *size = width * height * 32 / 8;
    ret = TRUE;
    GST_DEBUG_OBJECT (filter, "our frame size is %d bytes (%dx%d)", *size,
        width, height);
  }

  return ret;
}

static GstFlowReturn
gst_redgreentv_transform (GstBaseTransform * trans, GstBuffer * in, GstBuffer * out)
{
  GstRedGreenTV *redgreentv = GST_REDGREENTV (trans);
  int width = redgreentv->width;
  int height = redgreentv->height;
  guint8 *src = (guint8 *) GST_BUFFER_DATA (in);
  guint8 *dest = (guint8 *) GST_BUFFER_DATA (out);
  GstFlowReturn ret = GST_FLOW_OK;
  gint x, y;
  gdouble red, green;
  gdouble direction;

  gst_buffer_stamp (out, in);

  red = 0;
  green = 0;
  for (y = 0; y < height; y++) {
    for (x = 0; x < width; x++) {
      gdouble ratio = (gdouble) (src[0]) / (src[0] + src[1] + src[2]/8);
      if (ratio > 0.55)
        red+= ratio;
      else if (ratio < 0.45)
        green += (1.0 - ratio);
      src += 4;
    }
  }

  red = (gdouble) (red) / (width * height) * redgreentv->red_weight;
  green = (gdouble) (green) / (width * height) * redgreentv->green_weight;

  if (red < green) {
    direction = MIN (3 * (green - red), 100.0);
  } else {
    direction = MAX (3 * (green - red), -100.0);
  }
  src = (guint8 *) GST_BUFFER_DATA (in);
  if (direction > -30 && direction < 30) {
    memcpy (dest, src, sizeof (guint32) * width * height);
  }
  else {
    if (direction > 0) {
      for (y = 0; y < height; y++) {
        for (x = 0; x < width; x++) {
          *dest++ = 0;
          *dest++ = src[1];
          *dest++ = 0;
          *dest++ = src[3];
          src += 4;
        }
      }
    } else {
      for (y = 0; y < height; y++) {
        for (x = 0; x < width; x++) {
          *dest++ = src[0];
          *dest++ = 0;
          *dest++ = 0;
          *dest++ = src[3];
          src += 4;
        }
      }
    }
  }
  g_print ("red = %g, green = %g, direction %g\n", red, green, direction);

  if (redgreentv->direction != direction) {
    redgreentv->direction = direction;
    g_object_notify (G_OBJECT (trans), "direction");
  }


  return ret;
}

static void
gst_redgreentv_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstRedGreenTV *redgreentv = GST_REDGREENTV (object);
  switch (prop_id) {
    case ARG_RED_WEIGHT:
      redgreentv->red_weight = g_value_get_double (value);
      break;
    case ARG_GREEN_WEIGHT:
      redgreentv->red_weight = g_value_get_double (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_redgreentv_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  GstRedGreenTV *redgreentv = GST_REDGREENTV (object);
  switch (prop_id) {
    case ARG_RED_WEIGHT:
      g_value_set_double (value, redgreentv->red_weight);
      break;
    case ARG_GREEN_WEIGHT:
      g_value_set_double (value, redgreentv->green_weight);
      break;
    case ARG_DIRECTION:
      g_value_set_double (value, redgreentv->direction);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_redgreentv_base_init (gpointer g_class)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);

  gst_element_class_set_details (element_class, &redgreentv_details);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_redgreentv_sink_template));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_redgreentv_src_template));
}

static void
gst_redgreentv_class_init (gpointer klass, gpointer class_data)
{
  GObjectClass *gobject_class;
  GstElementClass *element_class;
  GstBaseTransformClass *trans_class;

  gobject_class = (GObjectClass *) klass;
  element_class = (GstElementClass *) klass;
  trans_class = (GstBaseTransformClass *) klass;

  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = gst_redgreentv_set_property;
  gobject_class->get_property = gst_redgreentv_get_property;

  trans_class->set_caps = GST_DEBUG_FUNCPTR (gst_redgreentv_set_caps);
  trans_class->get_unit_size = GST_DEBUG_FUNCPTR (gst_redgreentv_get_unit_size);
  trans_class->transform = GST_DEBUG_FUNCPTR (gst_redgreentv_transform);

  g_object_class_install_property (gobject_class,
      ARG_RED_WEIGHT,
      g_param_spec_double ("red-weight",
          "red-weight",
          "Weighting factor for ambient RED content",
          0, 1000.0, DEFAULT_RED_WEIGHT, G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
      ARG_GREEN_WEIGHT,
      g_param_spec_double ("green-weight",
          "green-weight",
          "Weighting factor for ambient GREEN content",
          0, 1000.0, DEFAULT_RED_WEIGHT, G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
      ARG_DIRECTION,
      g_param_spec_double ("direction",
          "direction",
          "Direction value from content, -100.0 for full red, 100.0 for full green",
          -100.0, 100.0, 0.0, G_PARAM_READABLE));

}

static void
gst_redgreentv_init (GTypeInstance * instance, gpointer g_class)
{
  GstRedGreenTV *redgreentv = GST_REDGREENTV (instance);
  redgreentv->red_weight = DEFAULT_RED_WEIGHT;
  redgreentv->green_weight = DEFAULT_GREEN_WEIGHT;
}

GType
gst_redgreentv_get_type (void)
{
  static GType redgreentv_type = 0;

  if (!redgreentv_type) {
    static const GTypeInfo redgreentv_info = {
      sizeof (GstRedGreenTVClass),
      gst_redgreentv_base_init,
      NULL,
      gst_redgreentv_class_init,
      NULL,
      NULL,
      sizeof (GstRedGreenTV),
      0,
      gst_redgreentv_init,
    };

    redgreentv_type = g_type_register_static (GST_TYPE_VIDEO_FILTER,
        "GstRedGreenTV", &redgreentv_info, 0);
  }
  return redgreentv_type;
}

static gboolean
plugin_init (GstPlugin * plugin)
{
  if (!gst_element_register (plugin, "redgreentv", GST_RANK_NONE,
      gst_redgreentv_get_type ()))
    return FALSE;

  return TRUE;
}

/* Normally defined by AS_VERSION in configure.ac - see gst-template */
#define PACKAGE "RedGreen TV plugin" 
#define PACKAGE_NAME "RedGreenTv Element"
#define PACKAGE_ORIGIN "http://lca2007.linux.org.au"

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "redgreen",
    "plugin for redgreen detection",
    plugin_init, "0.1.0", "LGPL", PACKAGE_NAME, PACKAGE_ORIGIN);
